import numpy as np


def set_scale(self):
    x1, y1 = self.swrc
    if self.log_x:
        self.min_x = self.min_x_log
        if hasattr(self, 'max_x'):
            self.max_x = max(self.max_x, max(x1) * 1.5)
        else:
            self.max_x = max(x1) * 1.5
    else:
        self.min_x = min(0, min(x1) * 0.85)
        try:
            self.max_x
        except AttributeError:
            self.max_x = max(x1) * 1.05
    self.min_y1 = 0
    if not hasattr(self, 'max_y1'):
        self.max_y1 = max(y1) * 1.15
    if not self.ht_only:
        x2, y2 = self.unsat
        if hasattr(self, 'min_y2'):
            self.min_y2 = min(self.min_y2, min(y2) * 0.2)
        else:
            self.min_y2 = min(y2) * 0.2
        self.max_y2 = max(y2) * self.max_y2_log


def add_curve(self):
    import math
    self.set_scale()
    color = self.color[self.line_num % len(self.color)]
    style = self.style[self.line_num % len(self.style)]
    if self.log_x:
        x = 2**np.linspace(math.log2(self.min_x),
                           math.log2(self.max_x), num=self.curve_smooth)
    else:
        x = np.linspace(self.min_x, self.max_x, num=self.curve_smooth)
    if self.ht_only:
        self.curves_ht.append({'data': (x, self.f_ht(
            self.fitted, x)), 'color': color, 'style': style, 'legend': self.line_legend})
    else:
        self.curves_ht.append({'data': (x, self.f_ht(self.p_ht(
            self.fitted), x)), 'color': color, 'style': style, 'legend': self.line_legend})
        self.curves_hk.append({'data': (x, self.f_hk(
            self.fitted, x)), 'color': color, 'style': style, 'legend': self.line_legend})
    self.line_num += 1


def clear_curves(self):
    self.curves_ht = []
    self.curves_hk = []
    self.line_num = 0


def h_0to1(self, data):
    if not self.fig_h_0to1:
        return data
    x, y = data
    x = np.where(x == 0, 1, x)
    return (x, y)


def plot(self):
    import matplotlib.pyplot as plt  # type: ignore
    import matplotlib.ticker as ticker  # type: ignore

    if self.data_only:
        self.set_scale()

    # Set subplots
    if self.ht_only:
        fig, ax1 = plt.subplots(figsize=[self.fig_width, self.fig_height])
    else:
        fig, (ax1, ax2) = plt.subplots(2, figsize=[
            self.fig_width, self.fig_height_double])
    fig.subplots_adjust(
        top=1 - self.top_margin,
        bottom=self.bottom_margin,
        right=1 - self.right_margin,
        left=self.left_margin,
        hspace=self.hspace)

    # Draw plots, curves and legends
    ax1.plot(*self.h_0to1(self.swrc), color=self.color_marker, marker=self.marker,
             linestyle='', label=self.data_legend)
    if not self.data_only:
        if self.have_new_plot:
            self.add_curve()
        for curve in self.curves_ht:
            ax1.plot(*curve['data'], color=curve['color'],
                     linestyle=curve['style'], label=curve['legend'])
        if not self.ht_only:
            ax2.plot(*self.h_0to1(self.unsat), color=self.color_marker,
                     marker=self.marker, linestyle='', label='_nolegend_')
            for curve in self.curves_hk:
                ax2.plot(*curve['data'], color=curve['color'],
                         linestyle=curve['style'], label='_nolegend_')
        if hasattr(self, 'fp'):
            leg = fig.legend(loc=self.legend_loc, prop=self.fp,
                             facecolor=self.legend_facecolor)
        else:
            leg = fig.legend(
                loc=self.legend_loc, fontsize=self.legend_fontsize, facecolor=self.legend_facecolor)
        leg.get_frame().set_alpha(self.legend_opacity)

    # Draw scale
    if self.log_x:
        ax1.set_xscale("log")
    ax1.axis([self.min_x, self.max_x, self.min_y1, self.max_y1])
    if not self.ht_only:
        ax1.xaxis.set_major_formatter(ticker.NullFormatter())
        ax2.axis([self.min_x, self.max_x, self.min_y2, self.max_y2])
        if self.log_x:
            ax2.loglog(base=10)
        else:
            ax2.set_yscale = ("log")
            ax2.xaxis.set_major_formatter(ticker.ScalarFormatter())

    # Draw labels
    if self.ht_only:
        ax1.set_xlabel(self.label_head)
    else:
        ax2.set_xlabel(self.label_head)
    ax1.set_ylabel(self.label_theta)
    if not self.ht_only:
        ax2.set_ylabel(self.label_k)

    # Show and/or save figure
    if self.save_fig:
        plt.savefig(self.filename)
    if self.show_fig:
        plt.show()
    plt.close()



# Initialize and set models

def init_model(self):
    self.model = {}
    self.init_model_bc()
    self.init_model_vg()
    self.init_model_ln()
    self.init_model_fx()
    self.init_model_bc2()
    self.init_model_vg2()
    self.init_model_ln2()
    self.init_model_vgbc()
    self.init_model_kobc()
    self.init_model_fs()
    self.init_model_pe()
    self.output_format = {
        'qs': '.3f', 'qr': '.3f', 'qa': '.3f', 'w1': '.3f', 'a': '.3', 'a1': '.3', 'a2': '.3',
        'm': '.3f', 'n': '.3f', 'm1': '.3f', 'm2': '.3f', 'hm': '.2f', 'hm1': '.2f', 'hm2': '.2f',
        'sigma': '.3f', 'sigma1': '.3f', 'sigma2': '.3f', 'hb': '.2f', 'hb1': '.2f', 'hb2': '.2f',
        'hc': '.2f', 'l': '.3', 'l1': '.3', 'l2': '.3', 'he': '.2f',
        'Ks': '.2e', 'p': '.3f', 'p1': '.3f', 'p2': '.3f', 'q': '.3f', 'q1': '.3f', 'q2': '.3f', 'r': '.3f', 'omega': '.3'
    }
    self.r2_format = '.3f'


def set_model(self, model, const=[]):
    """Set model

    Parameters:

        model: name of the model
        const: constant parameters of the model. For example
               [[1,0.5], [2,0], [3,0.2]] sets 1st parameter as 0.5, 2nd parameter as 0 and 3rd parameter as 0.2
               [[1,0.5], 'qr=0']] sets 1st parameter as 0.5 and qr as 0
               [wrf, 'r=1'] where wrf is a tuple of all WRF parameters, sets the WRF parameters and r as 1
    """
    self.model_name = model
    self.f_ht, self.f_hk = self.model[model]['function']
    self.b_func = self.model[model]['bound']
    self.param = self.model[model]['param']
    self.model_k_only = self.model[model]['k-only']
    # get_init and get_wrf functions
    if 'get_init' in self.model[model]:
        self.get_init = self.model[model]['get_init']
    else:
        self.get_init = self.get_init_not_defined
    if 'get_wrf' in self.model[model]:
        self.get_wrf = self.model[model]['get_wrf']
    else:
        self.get_wrf = self.get_wrf_not_defined
    # Recostruct const to allow alternative expressions
    reconst = []
    for i in const:
        if '=' in str(i):  # expression like 'q=1'
            p, value = i.split('=')
            if p not in self.param:
                print('Parameter {0} not in this model'.format(p))
                exit(1)
            reconst.append([self.param.index(p) + 1, float(value)])
        elif len(i) > 2:  # tuple or list of all water retention parameters
            p = list(range(len(self.param)))
            for j in self.model_k_only:
                p.remove(j)
            if len(p) == len(i):
                k = 0
                for j in p:
                    reconst.append([j + 1, i[k]])
                    k += 1
            else:
                print(
                    '{0} parameters required for water retention function, but {1} parameters {2} given.'.format(len(p), len(i), i))
                exit()
        else:  # expression like [2, 0]
            reconst.append(i)
    self.const = sorted(reconst)
    # print(self.model_description)
    # Calculate self.p_k_only from self.model_k_only by eliminating constant
    # Note: when it is (0,1,3) where 2 is constant, it should be arranged to
    # (0,1,2)
    k_only = set(self.model_k_only)
    for c in sorted(self.const, reverse=True):
        if c[0] - 1 in sorted(k_only):
            k_only.remove(c[0] - 1)
        else:
            for i in sorted(k_only):
                if i > c[0] - 1:
                    k_only.remove(i)
                    k_only.add(i - 1)
    self.p_k_only = sorted(list(k_only), reverse=True)
    # Calculate self.const_ht by eliminating K-only parameters
    self.const_ht = []
    for c in self.const:
        if c[0] - 1 not in self.model_k_only:
            self.const_ht.append(
                [c[0] - sum(1 for x in self.model_k_only if x < c[0] - 1), c[1]])
    # Calculate self.param and self.param_ht
    self.param_ht = self.param
    for i in sorted(self.model_k_only, reverse=True):
        self.param_ht = self.param_ht[:i] + self.param_ht[i + 1:]
    for c in sorted(self.const_ht, reverse=True):
        self.param_ht = self.param_ht[:c[0] - 1] + self.param_ht[c[0]:]
    for c in sorted(self.const, reverse=True):
        self.param = self.param[:c[0] - 1] + self.param[c[0]:]
    # Model description
    self.param_const = []
    self.value_const = []
    for i in self.const:
        self.param_const.append(self.model[model]['param'][i[0] - 1])
        self.value_const.append(float(i[1]))
    self.const_description = self.format(
        self.param_const, ShowR2=False).format(*self.value_const)
    self.model_description = self.model_name + \
        ' model with ' + self.const_description


def get_init_not_defined(self):
    print('get_init function not defined for {0} model'.format(
        self.model_name))
    exit(1)


def get_wrf_not_defined(self):
    print('get_wrf function not defined for {0} model'.format(
        self.model_name))
    exit(1)



# Fredlund and Xing model
import numpy as np


def init_model_fx(self):
    self.model['FX'] = self.model['fx'] = {
        'function': (self.fx, False),
        'bound': self.bound_fx,
        'get_init': self.get_init_fx,
        'get_wrf': self.get_wrf_fx,
        'param': ['qs', 'qr', 'a', 'm', 'n'],
        'k-only': []
    }


def bound_fx(self):
    return [self.b_qs, self.b_qr, self.b_fxa, self.b_fxm, self.b_fxn]


def fx(self, p, x):
    p = list(p)
    for c in self.const_ht:
        p = p[:c[0] - 1] + [c[1]] + p[c[0] - 1:]
    return self.fx_se(p[2:], x) * (p[0] - p[1]) + p[1]


def fx_se(self, p, x):
    a, m, n = p
    return (np.log(np.e + (x / a)**n))**(-m)


def get_init_fx(self):  # a, m, n
    from .unsatfit import Fit
    x, t = self.swrc
    y = t / max(t)
    f = Fit()
    f.set_model('fx', const=[[1, 1], [2, 0]])
    f.swrc = (x, y)
    a, m = f.get_init_vg()
    n = 1 / (1 - m)
    vg = f.ini = (1 / a, 2.54 * (1 - 1 / n), 0.95 * n)
    f.optimize()
    if f.success:
        return f.fitted
    hm, sigma = f.get_init_ln()
    f.ini = (hm, 2.54, 1.52 / sigma)
    f.optimize()
    if f.success:
        return f.fitted
    hb, l = f.get_init_bc()
    n = l + 1
    f.ini = (hb, 2.54 * (1 - 1 / n), 0.95 * n)
    f.optimize()
    if f.success:
        return f.fitted
    f.set_model('fx', const=[[2, 0]])
    f.ini = (1, *vg)
    f.optimize()
    if f.success:
        return f.fitted[1:]
    return vg


def get_wrf_fx(self):
    from .unsatfit import Fit
    f = Fit()
    f.swrc = self.swrc
    f.debug = self.debug
    a, m, n = f.get_init_fx()
    f.set_model('fx', const=[])
    qs = max(f.swrc[1])
    f.ini = (qs, 0, a, m, n)
    f.optimize()
    if f.success:
        return f.fitted
    f.b_qs = (qs * 0.95, qs * 1.5)
    f.b_qr = (0, min(f.swrc[1]))
    f.optimize()
    if f.success:
        return f.fitted
    return f.ini



# VG1BC2 model
import numpy as np


def init_model_vgbc(self):
    self.model['VGBC'] = self.model['VG1BC2'] = self.model['VB'] = self.model['vgbc'] = {
        'function': (self.vgbc, self.vgbc_k),
        'bound': self.bound_vgbc,
        'get_init': self.get_init_vgbc,
        'get_wrf': self.get_wrf_vgbc,
        'param': ['qs', 'qr', 'w1', 'a1', 'm1', 'hb2', 'l2', 'Ks', 'p', 'q', 'r'],
        'k-only': [7, 8, 10]
    }
    self.model['VGBCIP'] = self.model['VG1BC2-IP'] = self.model['VB-IP'] = self.model['vgbcp2'] = {
        'function': (self.vgbc, self.vgbcp2_k),
        'bound': self.bound_vgbcp2,
        'get_init': self.get_init_vgbc,
        'get_wrf': self.get_wrf_vgbc,
        'param': ['qs', 'qr', 'w1', 'a1', 'm1', 'hb2', 'l2', 'Ks', 'p1', 'p2', 'q'],
        'k-only': [7, 8, 9]
    }
    self.model['VGBCCH'] = self.model['VG1BC2-CH'] = self.model['VBC'] = self.model['vgbcch'] = {
        'function': (self.vgbcch, self.vgbcch_k),
        'bound': self.bound_vgbcch,
        'get_init': self.get_init_vgbcch,
        'get_wrf': self.get_wrf_vgbcch,
        'param': ['qs', 'qr', 'w1', 'a1', 'm1', 'l2', 'Ks', 'p', 'q', 'r'],
        'k-only': [6, 7, 9]
    }
    self.model['VGBCCHIP'] = self.model['VG1BC2-CH-IP'] = self.model['VBC-IP'] = self.model['vgbcchp2'] = {
        'function': (self.vgbcchp2, self.vgbcchp2_k),
        'bound': self.bound_vgbcchp2,
        'get_init': self.get_init_vgbcch,
        'get_wrf': self.get_wrf_vgbcch,
        'param': ['qs', 'qr', 'w1', 'a1', 'm1', 'l2', 'Ks', 'p1', 'p2', 'q'],
        'k-only': [6, 7, 8]
    }


def bound_vgbc(self):
    return [self.b_qs, self.b_qr, self.b_w1, self.b_a1, self.b_m,
            self.b_hb2, self.b_lambda2, self.b_ks, self.b_p, self.b_q, self.b_r]


def vgbc(self, p, x):
    p = list(p)
    for c in self.const_ht:
        p = p[:c[0] - 1] + [c[1]] + p[c[0] - 1:]
    return self.vgbc_se(p, x) * (p[0] - p[1]) + p[1]


def vgbc_se(self, p, x):
    qs, qr, w, a1, m1, hb2, l2, q = p
    s1 = self.vg_se([a1, m1, q], x)
    s2 = np.where(x < hb2, 1, (x / hb2) ** (-l2))
    return w * s1 + (1 - w) * s2


def vgbc_k(self, p, x):
    par = list(p)
    for c in self.const:
        par = par[:c[0] - 1] + [c[1]] + par[c[0] - 1:]
    qs, qr, w, a1, m1, hb2, l2, ks, p, q, r = par
    w1b1 = w * (a1 ** q)
    w2b2 = (1 - w) / hb2 ** q / (q / l2 + 1)
    s1 = self.vg_se([a1, m1, q], x)
    s1 = 1 - (1 - s1**(1 / m1))**m1
    s2 = np.where(x < hb2, 1, (x / hb2) ** (-l2 - q))
    bunshi = w1b1 * s1 + w2b2 * s2
    bunbo = w1b1 + w2b2
    return ks * self.vgbc_se(par[:7] + [q], x)**p * (bunshi / bunbo)**r


def get_init_vgbc(self):  # w1, a1, m1, hb2, l2
    from .unsatfit import Fit
    n_max = 8
    x, t = self.swrc
    y = t / max(t)
    f = Fit()
    f.debug = self.debug
    f.swrc = (x, y)
    w, a, m1, l2 = f.get_init_vgbcch()
    f.set_model('vgbc', const=[[1, 1], [2, 0], 'q=1'])
    m_max = 1 - 1 / n_max
    if m1 > m_max:
        m1 = m_max - 0.001
    f.ini = (w, a, m1, 1 / a, l2)
    f.b_m = (0, m_max)
    f.optimize()
    if f.success:
        ch = f.fitted
        ch_r2 = f.r2_ht
    else:
        ch = f.ini
        ch_r2 = f.f_r2_ht(f.ini, x, y)
    if len(x) < 6:
        return ch
    swrc = list(zip(*f.swrc))
    swrc_sort = sorted(swrc, key=lambda x: x[0])
    swrc = list(zip(*swrc_sort))
    h = np.array(swrc[0])
    t = np.array(swrc[1])
    t_med = (max(t) + min(t)) / 2
    for i in range(len(h)):
        if t[i] < t_med:
            break
    if i < 3:
        i = 3
    if i > len(h) - 3:
        i = len(h) - 3
    f.swrc = (h[:i], t[:i] - t[i])
    try:
        a1, m1 = f.get_init_vg()
    except BaseException:
        return ch
    if m1 > m_max:
        m1 = m_max
    f.swrc = (h[i:], t[i:])
    try:
        hb2, l2 = f.get_init_bc()
    except BaseException:
        return ch
    f.set_model('vgbc', const=['qs=1', 'qr=0', 'q=1'])
    f.swrc = (h, t)
    f.ini = (1 - t[i], a1, m1, hb2, l2)
    f.optimize()
    if not f.success:
        return ch
    if ch_r2 > f.r2_ht:
        return ch
    return f.fitted


def get_wrf_vgbc(self):
    from .unsatfit import Fit
    n_max = 8
    f = Fit()
    f.swrc = self.swrc
    f.debug = self.debug
    w1, a1, m1, hb2, l2 = f.get_init_vgbc()
    f.set_model('vgbc', const=['qr=0', 'q=1'])
    f.ini = (max(f.swrc[1]), w1, a1, m1, hb2, l2)
    m_max = 1 - 1 / n_max
    f.b_m = (0, m_max)
    f.optimize()
    if f.success:
        return (f.fitted[0], 0, *f.fitted[1:], 1)
    return (f.ini[0], 0, *f.ini[1:], 1)

# VG1BC2 model with r=1 and independent p1, p2


def bound_vgbcp2(self):
    return [self.b_qs, self.b_qr, self.b_w1, self.b_a1, self.b_m,
            self.b_hb2, self.b_lambda2, self.b_ks, self.b_p, self.b_p, self.b_q]


def vgbcp2_k(self, p, x):
    par = list(p)
    for c in self.const:
        par = par[:c[0] - 1] + [c[1]] + par[c[0] - 1:]
    qs, qr, w, a1, m1, hb2, l2, ks, p1, p2, q = par
    w1b1 = w * (a1 ** q)
    w2b2 = (1 - w) / (hb2 ** q) / (q / l2 + 1)
    s1 = self.vg_se([a1, m1, q], x)
    w1a1 = w1b1 * (1 - (1 - s1**(1 / m1))**m1)
    w2a2 = w2b2 * (np.where(x < hb2, 1, (x / hb2) ** (-l2 - q)))
    # Seki et al. (2021) eq.(19)
    # s2 = np.where(x < hb2, 1, (x/hb2) ** (-l2))
    # Corrected from Si(h) to Se(h)
    se = self.vgbc_se(par[:7] + [q], x)
    bunshi = se**p1 * w1a1 + se**p2 * w2a2
    bunbo = w1b1 + w2b2
    return ks * bunshi / bunbo

# VG1BC2-CH model


def bound_vgbcch(self):
    return [self.b_qs, self.b_qr, self.b_w1, self.b_a1, self.b_m,
            self.b_lambda2, self.b_ks, self.b_p, self.b_q, self.b_r]


def vgbcch(self, p, x):
    p = list(p)
    for c in self.const_ht:
        p = p[:c[0] - 1] + [c[1]] + p[c[0] - 1:]
    return self.vgbcch_se(p, x) * (p[0] - p[1]) + p[1]


def vgbcch_se(self, p, x):
    qs, qr, w, a1, m1, l2, q = p
    hb2 = 1 / a1
    s1 = self.vg_se([a1, m1, q], x)
    # Ignore runtime warning, because divide by zero is warned when x=0
    import warnings
    warnings.simplefilter('ignore', category=RuntimeWarning)
    s2 = np.where(x < hb2, 1, (x / hb2) ** (-l2))
    return w * s1 + (1 - w) * s2


def vgbcch_k(self, p, x):
    par = list(p)
    for c in self.const:
        par = par[:c[0] - 1] + [c[1]] + par[c[0] - 1:]
    qs, qr, w, a1, m1, l2, ks, p, q, r = par
    hb2 = 1 / a1
    w1b1 = w * (a1 ** q)
    w2b2 = (1 - w) / hb2 ** q / (q / l2 + 1)
    s1 = self.vg_se([a1, m1, q], x)
    s1 = 1 - (1 - s1**(1 / m1))**m1
    s2 = np.where(x < hb2, 1, (x / hb2) ** (-l2 - q))
    bunshi = w1b1 * s1 + w2b2 * s2
    bunbo = w1b1 + w2b2
    return ks * self.vgbcch_se(par[:6] + [q], x)**p * (bunshi / bunbo)**r


def get_init_vgbcch(self):  # w, alpha, m1, l2
    from .unsatfit import Fit
    x, t = self.swrc
    y = t / max(t)
    f = Fit()
    f.debug = self.debug
    f.swrc = (x, y)
    w, a, m1, m2 = f.get_init_vg2ch()
    f.set_model('vgbcch', const=[[1, 1], [2, 0], 'q=1'])
    n2 = 1 / (1 - m2)
    if n2 < 1.1:
        n2 = 1.1
    l2 = n2 - 1
    if l2 > 3:
        l2 = 2.99
    f.ini = (w, a, m1, l2)
    f.b_lambda2 = (0.001, 3)
    f.optimize()
    if f.success:
        return f.fitted
    hb, hc, l1, l2 = f.get_init_bc2()
    w = 1 / (1 + (hc / hb)**(l2 - l1))
    n1 = l1 + 1
    f.ini = (w, 1 / hb, 1 - 1 / n1, l2)
    f.optimize()
    if f.success:
        return f.fitted
    f.ini = (w, 1 / hb, m1, l2)
    f.optimize()
    if f.success:
        return f.fitted
    a, m = f.get_init_vg()
    f.b_w1 = (0.8, 1)
    f.ini = (0.99, a, m, 0.001)
    f.optimize()
    if f.success:
        return f.fitted
    return f.ini


def get_wrf_vgbcch(self):
    from .unsatfit import Fit
    f = Fit()
    f.swrc = self.swrc
    f.debug = self.debug
    w, alpha, m1, l2 = f.get_init_vgbcch()
    f.set_model('vgbcch', const=['qr=0', 'q=1'])
    f.ini = (max(f.swrc[1]), w, alpha, m1, l2)
    f.optimize()
    if f.success:
        return (f.fitted[0], 0, *f.fitted[1:], 1)
    return (f.ini[0], 0, *f.ini[1:], 1)


# VG1BC2-CH model with r=1 and independent p1, p2


def bound_vgbcchp2(self):
    return [self.b_qs, self.b_qr, self.b_w1, self.b_a1, self.b_m,
            self.b_lambda2, self.b_ks, self.b_p, self.b_p, self.b_q]


def vgbcchp2(self, p, x):
    p = list(p)
    for c in self.const_ht:
        p = p[:c[0] - 1] + [c[1]] + p[c[0] - 1:]
    return self.vgbcch_se(p, x) * (p[0] - p[1]) + p[1]


def vgbcchp2_k(self, p, x):
    par = list(p)
    for c in self.const:
        par = par[:c[0] - 1] + [c[1]] + par[c[0] - 1:]
    qs, qr, w, a1, m1, l2, ks, p1, p2, q = par
    hb2 = 1 / a1
    w1b1 = w * (a1 ** q)
    w2b2 = (1 - w) / (hb2 ** q) / (q / l2 + 1)
    s1 = self.vg_se([a1, m1, q], x)
    w1a1 = w1b1 * (1 - (1 - s1**(1 / m1))**m1)
    w2a2 = w2b2 * (np.where(x < hb2, 1, (x / hb2) ** (-l2 - q)))
    # Seki et al. (2021) eq.(19)
    # s2 = np.where(x < hb2, 1, (x/hb2) ** (-l2))
    # bunshi = s1**p1 * w1a1 + s2**p2 * w2a2
    # Corrected from Si(h) to Se(h)
    se = self.vgbcch_se(par[:6] + [q], x)
    bunshi = se**p1 * w1a1 + se**p2 * w2a2
    bunbo = w1b1 + w2b2
    return ks * bunshi / bunbo



# Kosugi model
import numpy as np


def init_model_ln(self):
    self.model['Kosugi'] = self.model['KO'] = self.model['ln'] = {
        'function': (self.ln, self.ln_k),
        'bound': self.bound_ln,
        'get_init': self.get_init_ln,
        'get_wrf': self.get_wrf_ln,
        'param': ['qs', 'qr', 'hm', 'sigma', 'Ks', 'p', 'q', 'r'],
        'k-only': [4, 5, 6, 7],
    }


def bound_ln(self):
    return [self.b_qs, self.b_qr, self.b_hm, self.b_sigma,
            self.b_ks, self.b_p, self.b_q, self.b_r]


def ln(self, p, x):
    p = list(p)
    for c in self.const_ht:
        p = p[:c[0] - 1] + [c[1]] + p[c[0] - 1:]
    qs, qr, hm, s = p
    return self.ln_se([hm, s], x) * (qs - qr) + qr


def ln_se(self, p, x):
    from scipy.stats import norm  # type: ignore
    # Ignore runtime warning, because divide by zero is warned when x=0
    import warnings
    warnings.simplefilter('ignore', category=RuntimeWarning)
    return 1 - norm.cdf(np.log(x / p[0]) / p[1])


def ln_k(self, p, x):
    from scipy.stats import norm
    par = list(p)
    for c in self.const:
        par = par[:c[0] - 1] + [c[1]] + par[c[0] - 1:]
    qs, qr, hm, s, ks, p, q, r = par
    qq = 1 - norm.cdf(np.log(x / hm) / s + q * s)
    return ks * self.ln_se([hm, s], x)**p * qq**r


def get_init_ln(self):  # hm and sigma
    from .unsatfit import Fit
    x, t = self.swrc
    y = t / max(t)
    f = Fit()
    f.set_model('ln', const=[[1, 1], [2, 0]])
    f.swrc = (x, y)
    a, m = f.get_init_vg()
    n = 1 / (1 - m)
    s = 1.2 * (n - 1)**(-0.8)
    if s < 0.15:
        s = 0.15
    if s > 3:
        s = 3
    f.ini = (1 / a, s)
    f.b_sigma = (0, 3.2)
    f.optimize()
    return f.fitted


def get_wrf_ln(self):
    from .unsatfit import Fit
    f = Fit()
    f.swrc = self.swrc
    f.debug = self.debug
    hm, s = f.get_init_ln()
    f.set_model('ln', const=[])
    qs = max(f.swrc[1])
    if s > 2.4:
        s = 2.4
    f.ini = (qs, 0, hm, s)
    f.b_sigma = (0, 2.5)
    f.optimize()
    if f.success:
        f.b_sigma = (0, np.inf)
        import copy
        f2 = copy.deepcopy(f)
        f.optimize()
        if f.success:
            return f.fitted
        else:
            return f2.fitted
    hb, l = f.get_init_bc()
    sigma = 1.2 * l**(-0.8)
    if sigma > 2.5:
        sigma = 2.5
    f.ini = (qs, 0, hb, sigma)
    f.b_qs = (qs * 0.95, qs * 1.5)
    f.optimize()
    if f.success:
        return f.fitted
    return f.ini



# Peters model
import numpy as np


def init_model_pe(self):
    self.model['PK'] = self.model['Peters-KO'] = self.model['Peters'] = self.model['PE'] = self.model['pk'] = {
        'function': (self.pk, self.pk_k),
        'bound': self.bound_pk,
        'get_init': self.get_init_pk,
        'get_wrf': self.get_wrf_pk,
        'param': ['qs', 'qr', 'w1', 'hm', 'sigma1', 'he', 'Ks', 'p', 'a', 'omega'],
        'k-only': [6, 7, 8, 9]
    }


def bound_pk(self):
    return [self.b_qs, self.b_qr, self.b_w1, self.b_hm1, self.b_sigma,
            self.b_he, self.b_ks, self.b_p, self.b_a, self.b_w1]


def pk(self, p, x):
    p = list(p)
    for c in self.const_ht:
        p = p[:c[0] - 1] + [c[1]] + p[c[0] - 1:]
    return self.pk_se(p, x) * (p[0] - p[1]) + p[1]


def pk_se(self, p, x):
    qs, qr, w, ha, sigma1, he = p
    s1 = self.ln_se([ha, sigma1], x)
    bunbo = np.log(1 + he / ha)
    xm = 1 / (1 - np.log(2) / bunbo)
    s2 = np.where(x < ha, 1, xm * (1 - np.log(1 + x / ha) / bunbo))
    return w * s1 + (1 - w) * s2


def pk_k(self, p, x):
    from scipy.stats import norm
    par = list(p)
    for c in self.const:
        par = par[:c[0] - 1] + [c[1]] + par[c[0] - 1:]
    qs, qr, w, ha, sigma1, he, ks, p, a, omega = par
    s1 = 1 - norm.cdf(np.log(x / ha) / sigma1 + sigma1)
    k1 = self.pk_se(par[:6], x)**p * s1**2
    k2 = np.where(x < ha, 1, (x / ha) ** (-a))
    return ks * ((1 - omega) * k1 + omega * k2)


def get_init_pk(self, he):  # w1, hm, s1
    from .unsatfit import Fit
    x, t = self.swrc
    y = t / max(t)
    f = Fit()
    f.debug = self.debug
    f.swrc = (x, y)
    w1, h1, s1, l2 = f.get_init_kobcch()
    result = f.ini = (w1, h1, s1)
    f.set_model('Peters', const=[[1, 1], [2, 0], [6, he]])
    r2 = -1000
    f.b_sigma = (0, 3)
    for f.b_hm1 in [(h1 * 0.9, h1 * 1.1), (0, np.inf)]:
        f.optimize()
        if f.success and f.r2_ht > r2:
            result = f.fitted
            r2 = f.r2_ht
    if r2 < 0.8:
        f.b_sigma = (0, 5)
        hm, sigma = f.get_init_ln()
        if sigma > 4:
            sigma = 4
        w1, a, m1, l2 = f.get_init_vgbcch()
        n1 = 1 / (1 - m1)
        sigma1 = 1.2 * (n1 - 1)**(-0.8)
        if sigma1 > 3:
            sigma1 = 3
        f.ini = [(0.05, 0.95), (h1, hm, 1 / a), (sigma1, sigma, s1)]
        f.optimize()
        if f.success and f.r2_ht > r2:
            result = f.fitted
            r2 = f.r2_ht
    return result


def get_wrf_pk(self, he):
    from .unsatfit import Fit
    f = Fit()
    f.swrc = self.swrc
    f.debug = self.debug
    w1, hm, s1 = f.get_init_pk(he)
    if s1 > 2.5:
        s1 = 2.5
    f.set_model('pk', const=['qr=0', [6, he]])
    qs = max(f.swrc[1])
    f.ini = (qs, w1, hm, s1)
    f.b_sigma = (0, 2.5)
    f.b_qs = (qs * 0.95, qs * 3)
    f.optimize()
    if f.success:
        import copy
        f2 = copy.deepcopy(f)
        f.b_sigma = (0, 3)
        f.optimize()
        if f.success and f.r2_ht > f2.r2_ht:
            return (f.fitted[0], 0, *f.fitted[1:], he)
        else:
            return (f2.fitted[0], 0, *f2.fitted[1:], he)
    f.b_qs = (qs * 0.95, qs * 1.2)
    f.optimize()
    if f.success:
        return (f.fitted[0], 0, *f.fitted[1:], he)
    return (qs, 0, *f.ini[1:], he)



# Pore size distribution


def f_pore(self, p, r, C=0.149, epsilon=1e-8):
    """
    Pore-size distribution function dθ/dr

    Note that figure should be drawn with normal scale.
    For using log-scale of pore-size, use f_pore_log.

    Parameters:
    -----------
    p : dict or array-like
        parameter of water retention function
    r : numpy.ndarray
        pore size
    C : float
        constant for relationship h = C/r
        unit dependent. 0.149 cm^2
    epsilon: float
        increment of numerical differentiation

    Returns:
    --------
    numpy.ndarray
        dθ/dr
    """
    h = C / r
    return -C / (r**2) * self.f_dtdh(p, h, epsilon=epsilon)


def f_pore_log(self, p, r, C=0.149, epsilon=1e-8):
    """
    Pore-size distribution function dθ/d(ln r)

    Parameters:
    -----------
    p : dict or array-like
        parameter of water retention function
    r : numpy.ndarray
        pore size
    C : float
        constant for relationship h = C/r
        unit dependent. 0.149 cm^2
    epsilon: float
        increment of numerical differentiation

    Returns:
    --------
    numpy.ndarray
        dθ/d(ln r)
    """
    h = C / r
    return -h * self.f_dtdh(p, h, epsilon=epsilon)


def f_dtdh(self, p, h, epsilon=1e-8):
    """
    dθ/dh calculated with numerical differentiation,
            using 4-point central difference formula

    Parameters:
    -----------
    p : dict or array-like
        parameter of water retention function
    h : numpy.ndarray
        pressure head
    epsilon: float
        increment of numerical differentiation

    Returns:
    --------
    numpy.ndarray
        dθ/dh
    """
    theta_values = [self.f_ht(p, h + k * epsilon) for k in [-2, -1, 1, 2]]
    return sum(
        k * theta for k, theta in zip([1, -8, 8, -1], theta_values)) / (12 * epsilon)



def init_lsq(self):
    """Set initial parameters for least square optimization

    Parameters are passed to scipy.optimize.least_squares
    See description at
    https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.least_squares.html
    """
    self.lsq_method = 'trf'  # 'trf' and 'dogbox' can be selected. 'lm' is not possible as it has no bounds
    self.lsq_jac = '2-point'  # 2-point’, ‘3-point’, ‘cs’
    # Loss function. ‘linear’, ‘soft_l1’, ‘huber’, ‘cauchy', ‘arctan’
    self.lsq_loss = 'linear'
    self.lsq_verbose = 0  # Level of algorithm’s verbosity. (0, 1, 2)
    # ftol changes stepwise in this order
    self.lsq_ftol = [0.1, 0.01, 1e-3, 1e-4, 1e-6, 1e-8]
    # ftol for initial estimation of mutiple initial parameter sets
    self.lsq_ftol_global = [1, 0.1]
    self.lsq_max_nfev = 1000  # Number of evaluation



# Cost function
import numpy as np


def residual_ht(self, p, x, y):
    return self.f_ht(p, x) - y


def residual_ln_hk(self, p, x, y):
    return np.log(self.f_hk(p, x) / y)


def residual_log10_hk(self, p, x, y):
    return np.log10(self.f_hk(p, x) / y)


def f_r2_ht(self, p, x, y):
    mse_ht = np.average(self.residual_ht(p, x, y)**2)
    return 1 - mse_ht / self.var_theta


def f_r2_ln_hk(self, p, x, y):
    mse_ln_hk = np.average(self.residual_ln_hk(p, x, y)**2)
    return 1 - mse_ln_hk / self.var_ln_k


def p_ht(self, p):
    p = list(p)
    for c in self.p_k_only:
        if c + 2 > len(p):
            p = p[:c]
        else:
            p = p[:c] + p[c + 1:]
    return p


def total_cost(self, p, x1, y1, x2, y2):
    r2 = self.f_r2_ht(self.p_ht(p), x1, y1) + self.f_r2_ln_hk(p, x2, y2)
    return 2 - r2



# dual-VG model
import numpy as np


def init_model_vg2(self):
    self.model['DV'] = self.model['dual-VG'] = self.model['vg2'] = {
        'function': (self.vg2, self.vg2_k),
        'bound': self.bound_vg2,
        'get_init': self.get_init_vg2,
        'get_wrf': self.get_wrf_vg2,
        'param': ['qs', 'qr', 'w1', 'a1', 'm1', 'a2', 'm2', 'Ks', 'p', 'q', 'r'],
        'k-only': [7, 8, 10]
    }
    self.model['DVCH'] = self.model['DVC'] = self.model['dual-VG-CH'] = self.model['vg2ch'] = {
        'function': (self.vg2ch, self.vg2ch_k),
        'bound': self.bound_vg2ch,
        'get_init': self.get_init_vg2ch,
        'get_wrf': self.get_wrf_vg2ch,
        'param': ['qs', 'qr', 'w1', 'a1', 'm1', 'm2', 'Ks', 'p', 'q', 'r'],
        'k-only': [6, 7, 9]
    }
    self.model['vg2chca'] = {
        'function': (self.vg2ch, self.vg2chca_k),
        'bound': self.bound_vg2ch,
        'get_init': self.get_init_vg2ch,
        'param': ['qs', 'qr', 'w1', 'a1', 'm1', 'm2', 'Ks', 'p', 'q', 'a'],
        'k-only': [6, 7, 9]
    }


def bound_vg2(self):
    return [self.b_qs, self.b_qr, self.b_w1, self.b_a1, self.b_m,
            self.b_a2, self.b_m, self.b_ks, self.b_p, self.b_q, self.b_r]


def vg2(self, p, x):
    p = list(p)
    for c in self.const_ht:
        p = p[:c[0] - 1] + [c[1]] + p[c[0] - 1:]
    return self.vg2_se(p, x) * (p[0] - p[1]) + p[1]


def vg2_se(self, p, x):
    s1 = self.vg_se([p[3], p[4], p[7]], x)
    s2 = self.vg_se([p[5], p[6], p[7]], x)
    return p[2] * s1 + (1 - p[2]) * s2


def vg2_k(self, p, x):
    par = list(p)
    for c in self.const:
        par = par[:c[0] - 1] + [c[1]] + par[c[0] - 1:]
    qs, qr, w, a1, m1, a2, m2, ks, p, q, r = par
    w1a1q = w * (a1 ** q)
    w2a2q = (1 - w) * (a2 ** q)
    s1 = self.vg_se([a1, m1, q], x)
    s2 = self.vg_se([a2, m2, q], x)
    bunshi = w1a1q * (1 - (1 - s1**(1 / m1))**m1) + \
        w2a2q * (1 - (1 - s2**(1 / m2))**m2)
    bunbo = w1a1q + w2a2q
    return ks * self.vg2_se(par[:7] + [q], x)**p * (bunshi / bunbo)**r


def get_init_vg2(self):  # w, alpha1, m1, alpha2, m2
    from .unsatfit import Fit
    n_max = 8
    x, t = self.swrc
    y = t / max(t)
    f = Fit()
    f.debug = self.debug
    f.swrc = (x, y)
    w, a, m1, m2 = f.get_init_vg2ch()
    f.set_model('vg2', const=[[1, 1], [2, 0], [10, 1]])
    f.ini = (w, a, m1, a, m2)
    m_max = 1 - 1 / n_max
    f.b_m = (0, m_max)
    f.optimize()
    if f.success:
        ch = f.fitted
        ch_r2 = f.r2_ht
    else:
        ch = f.ini
        ch_r2 = f.f_r2_ht(f.ini, x, y)
    if len(x) < 6:
        return ch
    swrc = list(zip(*f.swrc))
    swrc_sort = sorted(swrc, key=lambda x: x[0])
    swrc = list(zip(*swrc_sort))
    h = np.array(swrc[0])
    t = np.array(swrc[1])
    t_med = (max(t) + min(t)) / 2
    for i in range(len(h)):
        if t[i] < t_med:
            break
    if i < 3:
        i = 3
    if i > len(h) - 3:
        i = len(h) - 3
    f.swrc = (h[:i], t[:i] - t[i])
    try:
        a1, m1 = f.get_init_vg()
    except BaseException:
        return ch
    if m1 > m_max:
        m1 = m_max
    f.swrc = (h[i:], t[i:])
    try:
        a2, m2 = f.get_init_vg()
    except BaseException:
        return ch
    if m2 > m_max:
        m2 = m_max
    f.set_model('DV', const=['qs=1', 'qr=0', 'q=1'])
    f.swrc = (h, t)
    f.ini = (1 - t[i], a1, m1, a2, m2)
    f.optimize()
    if not f.success:
        return ch
    if ch_r2 > f.r2_ht:
        return ch
    return f.fitted


def get_wrf_vg2(self):
    from .unsatfit import Fit
    n_max = 8
    f = Fit()
    f.swrc = self.swrc
    f.debug = self.debug
    w, a1, m1, a2, m2 = f.get_init_vg2()
    f.set_model('vg2', const=['qr=0', 'q=1'])
    f.ini = (max(f.swrc[1]), w, a1, m1, a2, m2)
    m_max = 1 - 1 / n_max
    f.b_m = (0, m_max)
    f.optimize()
    if f.success:
        return (f.fitted[0], 0, *f.fitted[1:], 1)
    return (f.ini[0], 0, *f.ini[1:], 1)

# dual-VG-CH model


def bound_vg2ch(self):
    return [self.b_qs, self.b_qr, self.b_w1, self.b_a1, self.b_m,
            self.b_m, self.b_ks, self.b_p, self.b_q, self.b_r]


def vg2ch(self, p, x):
    p = list(p)
    for c in self.const_ht:
        p = p[:c[0] - 1] + [c[1]] + p[c[0] - 1:]
    return self.vg2ch_se(p, x) * (p[0] - p[1]) + p[1]


def vg2ch_se(self, p, x):
    s1 = self.vg_se([p[3], p[4], p[6]], x)
    # s2 = self.vg_se([p[4], p[5], p[6]], x)
    s2 = self.vg_se([p[3], p[5], p[6]], x)  # Fixed
    return p[2] * s1 + (1 - p[2]) * s2


def vg2ch_k(self, p, x):
    par = list(p)
    for c in self.const:
        par = par[:c[0] - 1] + [c[1]] + par[c[0] - 1:]
    qs, qr, w, a1, m1, m2, ks, p, q, r = par
    s1 = self.vg_se([a1, m1, q], x)
    s2 = self.vg_se([a1, m2, q], x)
    bunshi = w * (1 - (1 - s1**(1 / m1))**m1) + \
        (1 - w) * (1 - (1 - s2**(1 / m2))**m2)
    return ks * self.vg2ch_se(par[:6] + [q], x)**p * bunshi**r


def get_init_vg2ch(self):  # w, alpha, m1, m2
    from .unsatfit import Fit
    n_max = 8
    x, t = self.swrc
    y = t / max(t)
    f = Fit()
    f.debug = self.debug
    f.swrc = (x, y)
    m_max = 1 - 1 / n_max
    f.b_m = (0, m_max)
    a, m1 = f.get_init_vg()
    if m1 > m_max:
        m1 = m_max - 0.01
    hb, l = f.get_init_bc()
    hc, max_hc = hb * (0.3, 0.1) ** (-1 / l)
    w = 1 / (1 + (hc / hb)**(-l))
    i = sum(x < max_hc)
    if len(x) - i > 1:
        x = np.log(x[i:] / hb)
        y = -np.log(y[i:] / (1 - w))
        l2 = self.linear_regress(x, y)
        if l2 < 0.01:
            l2 = 0.01
    else:
        l2 = 0
    n2 = l2 + 1
    m2 = 1 - 1 / n2
    if m2 > m_max:
        m2 = m_max - 0.01
    f.set_model('vg2ch', const=[[1, 1], [2, 0], [
                3, w], [4, a], [6, m2], [9, 1]])
    f.ini = (m1)
    f.optimize()
    m1, = f.fitted
    f.set_model('vg2ch', const=[[1, 1], [2, 0], [9, 1]])
    f.ini = (w, a, m1, m2)
    f.optimize()
    if f.success:
        return f.fitted
    a, m = f.get_init_vg()
    f.ini = (0.5, a, m, m)
    f.optimize()
    if f.success:
        return f.fitted
    return f.ini


def get_wrf_vg2ch(self):
    from .unsatfit import Fit
    f = Fit()
    f.swrc = self.swrc
    f.debug = self.debug
    w, a, m1, m2 = f.get_init_vg2ch()
    f.set_model('vg2ch', const=['qr=0', 'q=1'])
    f.ini = (max(f.swrc[1]), w, a, m1, m2)
    f.optimize()
    if f.success:
        return (f.fitted[0], 0, *f.fitted[1:], 1)
    return (f.ini[0], 0, *f.ini[1:], 1)

# dual-VG-CH constant a model


def vg2chca_k(self, p, x):
    par = list(p)
    for c in self.const:
        par = par[:c[0] - 1] + [c[1]] + par[c[0] - 1:]
    qs, qr, w, a1, m1, m2, ks, p, q, a = par
    n2 = q / (1 - m2)
    r = (a + p * q) / n2 - p
    s1 = self.vg_se([a1, m1, q], x)
    s2 = self.vg_se([a1, m2, q], x)
    bunshi = w * (1 - (1 - s1**(1 / m1))**m1) + \
        (1 - w) * (1 - (1 - s2**(1 / m2))**m2)
    return ks * self.vg2ch_se(par[:6] + [q], x)**p * bunshi**r
